Program TInputDemo;
{
  YA-QGIR v0.2.0 - (C)1996 Eddy L O Jansson / DFR Research & Engineering

 Description:

  Yet another quite general input routine. After using my old version for
  many years it was time to rewrite it from the ground up. This is the result.
  The major difference is the 'edit-window' support. You can edit string
  up to 255 characters in a 'window' of any size. The code is quite clean,
  so you should have no problems adding or removing features.

 Usage:

  Instead of passing lot's of parameters you pass a record. If you rather
  pass parameters you can easily encapsule the TInput() routine.

  Setup a TInputRec record. Call TInput() with that records variable. Done.

 Notes:

  This demonstration and TInput() itself uses some routines from my
  my STMR_LOW and STMR_MID units. If you don't have these, search the
  SWAG or simply rewrite the needed functions yourself. Or contact me.

 Copyright:

  This is FREEWARE. Use it, modify it, spread it. If you want you can send
  me a 'thank you' e-mail, or maybe a free copy of whatever you've made.

  E-mail: eddy@saru.ct.se     FidoNET: 2:206/233     KlopperNET: 204:16/103
}
Uses STMR_LOW,STMR_MID,STMR_STR,ASMUtils;


TYPE
 TInputRec = Record { 267 bytes. Increase string length as needed. }
  X,Y         :Byte;         { Position of input }
  S           :String[80];   { Outstring (result) }
  StartS      :String[80];   { Default string }
  Width       :Byte;         { Width of edit-window }
  MaxLen      :Byte;         { Maximum string-length allowed to be input }
  StartPos    :Byte;         { Where to start edit. -1 = end of string }
  BackChar    :Char;         { Background character. #176 is nice. }
  Arrows,                    { Use 'direction' arrows? }
  Insert      :Boolean;      { Insert or Overwrite? }
  AcceptSet   :Set of Char;  { Standard keys accepted as input }
  NormExitSet :Set of Char;  { Standard keys to exit on }
  ExtExitSet  :Set of Char;  { Extended keys to exit on }
  Status      :Word;        { Returns the key used to terminate the routine }
                           { If extended key then status = $00xx else $xx00 }
 end;

VAR
 I           :TInputRec;
 XX,YY       :Byte;

{---------------------------------------------------------------------------}

Procedure TInput(var TI: TInputRec);
{ This is the /static/ version. Consider a 'UpdateIfKeypressed' version }
var
 Done           :Boolean; { Exit routine if TRUE }
 P              :Byte;    { Pointer into the string }
 LP             :Byte;    { Pointer into pos of string in leftmost pos of window }
 Ch,ChE         :Char;
 VP             :Integer; { Cursorpos in window (0 to TI.Width-1 }
 S,SS           :String;  { local temp-variables }

 Procedure UpdateWindow;
 begin
  SS:=Copy(S,LP,TI.Width); { Original: Copy_P(S,LP,TI.Width,SS); }
  PadRight_P(SS,Ti.Width,Ord(TI.BackChar));
  WriteXY(TI.X,TI.Y,SS);
  if TI.Arrows then begin
   if LP>1 then WriteXY(TI.X-1,TI.Y,#17) else WriteXY(TI.X-1,TI.Y,#32);
   if Length(S)>=LP+TI.Width then WriteXY(TI.X+TI.Width,TI.Y,#16) else WriteXY(TI.X+TI.Width,TI.Y,#32);
  end;
  GotoXY(TI.X+VP,TI.Y);
 end;

 Procedure MoveCursLeft;
 begin
  Dec(P); Dec(VP);
  if VP<0 then begin Dec(LP); VP:=0; end;
  UpdateWindow;
 end; { LeftKey }

 Procedure MoveCursRight;
 begin
  if P+1<TI.MaxLen then
   begin
    Inc(P); Inc(VP);
    if VP>TI.Width-1 then begin Inc(LP); VP:=TI.Width-1; end;
   end;
  UpdateWindow;
 end; { RightKey }

 Procedure doExtendedKey;
 begin
  ChE:=ReadKey; { Read Extended Key }

  if NOT (ChE in TI.ExtExitSet) then begin

  { Leftkey }
  if (ChE=#75) AND (P>0) then MoveCursLeft;
  { Rightkey }
  if (ChE=#77) AND (P<Length(S)) then MoveCursRight;
  { Insert/Overwrite }
  if (ChE=#82) then
   begin
    TI.Insert:=NOT TI.Insert;
    if TI.Insert then NormalCursor else BlockCursor;
   end;
  { Home }
  if (ChE=#71) then
   begin
    P:=0; VP:=0; LP:=1;
    UpdateWindow;
   end;
  { End }
  if (ChE=#79) then
   begin
    P:=Length(S);
    VP:=MinI(Length(S),TI.Width-1);
    LP:=MaxI(1,P-TI.Width+2);
    UpdateWindow;
   end;
  { Delete }
  if (ChE=#83) then
   begin
    Delete(S,P+1,1);
    UpdateWindow;
   end;
  end else begin Done:=True; TI.Status:=Ord(ChE); end;
 end; { doExtendedKey }

 Procedure doNormalKey;
 begin
  if Length(S)<TI.MaxLen then
  begin
   if TI.Insert then begin
    Insert(Ch,S,P+1);
    MoveCursRight;
   end else begin { if overwrite }
    if P+1>Length(S) then S:=S+Ch else S[P+1]:=Ch;
    MoveCursRight;
   end;
  end; { if not string too long }
 end; { doNormalKey }

begin
 { Fix background character }
 if TI.BackChar='' then TI.BackChar:=' ';
 { Don't allow entering of strings longer than supported by the structure }
 if TI.MaxLen>SizeOf(TI.S)-1 then TI.MaxLen:=SizeOf(TI.S)-1;
 { Truncate string if longer than allowed }
 if Length(TI.StartS) > TI.MaxLen then Left(TI.StartS,TI.MaxLen);
 { Fix startpos if larger than string }
 if TI.StartPos > Length(TI.StartS) then TI.StartPos:=Length(TI.StartS);
 { Setup local edition of TI.startS }
 S:=TI.StartS;
 { Write the 'entry-form' }
 WriteXY(TI.X,TI.Y,RepeatChar(TI.BackChar,TI.Width));
 { Setup a most important variable, the pointer into the string }
 P:=TI.StartPos;
 { Setup VP (cursorpos) }

 VP:=MinI(Length(S),TI.Width);

 { Setup LP - The start-pos of the string in the window }
 LP:=MaxI(1,P-TI.Width+1); { +1 because we want to start _behind_ the string }
 if TI.Insert then NormalCursor else BlockCursor;
 { Update the entry-window }
 UpdateWindow;
 Done:=False;
 REPEAT
  Ch:=ReadKey;
  if Ch in TI.NormExitSet then
   begin
    Done:=True;
    TI.Status:=Word(Ord(Ch) SHL 8);
   end else begin
    if Ch=#0 then doExtendedKey else
     if Ch in TI.AcceptSet then doNormalKey else
      begin { Very special keys }
       { Backspace }
       if (Ch=#8) and (P>0) then
        begin
         Delete(S,P,1);
         MoveCursLeft;
        end;
       { CTRL+Y }
       if (Ch=#25) then
        begin
         S:=''; VP:=0; P:=0; LP:=1;
         UpdateWindow;
        end;
      end; { .. was special-special key }

   end; { take care of character }
 UNTIL Done;

 { Remove arrows }
 if TI.Arrows then begin
  WriteXY(TI.X-1,TI.Y,#32);
  WriteXY(TI.X+TI.Width,TI.Y,#32);
 end;

 TI.S:=S;
end;

{ A simple demonstration }
BEGIN
 STMR_LOW.Init;
 ClrScr;

 XX:=10;
 YY:=5;

 WindowBox2(XX,YY,60,5,127,112,127,SinFrame,0,62,' TInput() Example ',True);
 WriteXYC(XX+3,YY+2,112,'Please enter something:');

 With I do begin
  X:=XX+27;
  Y:=YY+2;
  StartS:='0123456789 --- ABCDEFGHIJ';
  S:='';
  Width:=20;
  MaxLen:=SizeOf(S)-1;
  StartPos:=80;
  BackChar:='';
  Insert:=True;
  Arrows:=True;
  AcceptSet  :=[#32..#254];
  NormExitSet:=[#27,#13];
  ExtExitSet :=[#72,#80];
 end;

 { Setup arrow colors. Note: Arrows are placed -outside- the edit window. }
 SetAttrXY(I.X-1,I.Y,113);
 SetAttrXY(I.X+I.Width,I.Y,113);

 { Do the input }
 TInput(I);
 WriteXYCode(0,12,'~'#15'You Entered ~'#7'"~'#27+I.S+'~'#7'"');
 GotoXY(0,SH-2);
END.

{

 Encapsuled example:

 Ut:=TInputA(1,1,'Hej igen!',8,16,I.Status));

Function TInputA(const X,Y: Byte;const StartS: String;const Width,MaxLen: Byte;
                 var InputStatus: Word): String;
var
 I              :TInputRec;
begin
 I.X:=X;
 I.Y:=Y;
 I.StartS:=StartS;
 I.Width:=Width;
 I.MaxLen:=MaxLen;
 With I do begin
  S:='';
  StartPos:=Length(StartS);
  BackChar:='';
  Insert:=True;
  Arrows:=True;
  AcceptSet  :=[#32..#255];
  NormExitSet:=[#27,#13];
  ExtExitSet :=[#72,#80];
 end;
 TInput(I);
 TInputA:=I.S;
end;
}
