{ ** FIX MULTIPAGE IN 'CLRSCR' }

Unit STMR_LOW; { Sarumans Text Mode Routines (Low-level) }

{
 STMR_LOW v0.2.0 - (C)1996, 97 Eddy L O Jansson / DFR Research & Engineering

 Description:

  Various lowlevel routines to manipulate textmodes. Supports extended
  (SVGA/VESA) textmodes (132*43, 132*25, etc), Virtual Screens,
  block manipulation and much more. This is the /low-level/ portion.

 Usage:

  Before you can use this unit you must call the Init(); procedure. I have
  chosen not to add this to the unit-init as that would add to the overhead
  of your executable if you include STMR_LOW but don't use any of it's
  routines. Current overhead is in the order of 48 bytes.

  SetScreen(3);      ;setup the screenmode you want and then...
  STMR_LOW.Init;     ;...initialize it. Now you can use the STMR library.

 Notes:

 * Monochrome support - haven't tested it.

 * Some routines will only work on 386+ processors. Easily fixed.

 * I've added the 'cld' instruction to all functions that use the assembly
   'string' functions. It can be removed if you know what you're doing.

 * BP/TP's CRT unit will reset the screenmode if it's anything but mode
   1-3 or 7. This means that you'll have to patch the RTL if you want
   your program to support whatever screen mode 'DOS' is currently in
   _OR_ you must not use CRT at all. I'd suggest you just drop CRT altogether.

 * Some of you may want the change the way I'm passing Attributes+Chars,
   as a word instead of as two separate bytes. You can assemble this word
   using the following code: Word(Attribute SHL 8)+Character

 Future:

 * I'm considering replacing the lookup-table (YTable) with standard muls.

 History:

  0.1.0  Inhouse use only.
  0.1.1  1:st official release (SWAG)
  0.2.0  Fixed TextBoxGet/Put() (broken in 0.1.1! Sorry 'bout that.)
         Added Attr2B(), Attr2F() and FB2Attr()
         Added multipage support.
         Fixed WhereXY()
         Added Attr2FB()
         Syntax changed: FillScreen()
         2:nd official release (SWAG)

 Copyright:

  This is FREEWARE. Use it, modify it, spread it. If you want you can send
  me a 'thank you' email, or maybe a free copy of whatever you've made.
  I beg you; please send me email if you use this code. It's a big "boost"
  to recieve email from people who use and enjoy ones code. Write a line or
  two about what you've made using this code. Make me happy. Thanks.

  e-mail: eddy@saru.ct.se     FidoNET: 2:206/233     KlopperNET: 204:16/103
}


{ the "MULTIPAGE" directive enables multipage support. Textmodes use
  little memory (height*width*2 bytes), but the videosegment is 64Kb
  nonetheless. This would be a complete waste if it weren't for the
  ability to use different display 'pages'. Pages are a kind of virtual
  screens; you can read/write to and from what would normally be off-display
  memory. This off-display memory can then be displayed at any time.
  The global variables #RDP# (Read Display Page) and #WDP# (Write Display Page)
  controls where in memory (to what page) the screenfunctions will read/write.

  You you don't plan to use this feature then disable it = Saves code/space.
}

{$DEFINE MULTIPAGE}

INTERFACE

const
 MaxScreenHeight        = 60;

{ Foreground and background color constants }

 Black                  = 0;
 Blue                   = 1;
 Green                  = 2;
 Cyan                   = 3;
 Red                    = 4;
 Magenta                = 5;
 Brown                  = 6;
 LightGray              = 7;
 DarkGray               = 8;
 LightBlue              = 9;
 LightGreen             = 10;
 LightCyan              = 11;
 LightRed               = 12;
 LightMagenta           = 13;
 Yellow                 = 14;
 White                  = 15;
 Blink                  = 128;

var
 SW,                              { Screen Width }
 SH,                              { Screen Height }
 {$IFDEF MULTIPAGE}
 VDP,                             { Visible/Current Display Page } { :Word Absolute $40:$62;}
 RDP,                             { Read Display Page }
 WDP,                             { Write Display Page }
 {$ENDIF}
 PageSize,                        { :Word Absolute $40:$4C;}
 VisiblePageSize        :Word;    { SW*SH*2 }



 CSM,                             { Current Screen Mode }
 PreviousVideoMode      :Byte;
 TextMem                :Pointer; { This is the baseadress for all r/w routines }
 CYTable                :Array[1..MaxScreenHeight] of Word;
{-[ Function Interface ]----------------------------------------------------}

Procedure Init;
{ This routine is used once when the unit is initalized and it should
  be used every time you change the textmode resolution. It updates some
  of the global variables. Especially the CYTable which MUST be up to
  date. Let's repeat this one more time. You *MUST* call this procedure
  after /any/ change in screen resolution.
}

Procedure SetScreen(const mode: byte);
{$IFDEF MULTIPAGE}
Procedure SetDisplayPage(const page: word);
{$ENDIF}
Function  SetVESAMode(const mode: Word): Byte;

Function  GetScreen: Byte;
Function  GetScreenWidth: Byte;
Function  GetScreenHeight: Byte;

Procedure ClrScr;
{
  Written: 96-09-21, Eddy L O Jansson (2:206/233)

  Description:

   Clears the screen by filling it with attribute 7 and spaces.

  Notes:

   Does not home the cursor.
   Will clear only page zero.

  See Also:

   #FillScreen#
}

Procedure FillScreen(const AttrChar :Word);
{
  Updated: 97-06-11 (multipage), 97-07-26 (syntax)
}

Procedure WriteXY(const X,Y: Word;const Uts: String);
{
  Written: 96-09-21, Eddy L O Jansson (2:206/233)
  Updated: 96-09-25, 96-11-23 (zerostring), 97-06-11 (multipage)

  Description:

  Write a string at X,Y. Coordinates are zerobased and the
  routine utilizes the CYTABLE array to enable it to work with
  different screenwidths at the fastest possible speed.

  Notes:

   Does not set the color.

  See Also:

   #WriteXYC# #WriteXYCode#
}

Procedure WriteXYC(const X,Y: Word;const Color: Byte;const Uts: String);
{
  Written: 96-09-21, Eddy L O Jansson (2:206/233)
  Updated: 96-09-25, 96-11-23 (zerostring), 97-06-11 (multipage)

  Description:

  Write a string at X,Y using Color. Coordinates are zerobased and
  the routine utilizes the CYTABLE array to enable it to work with
  different screenwidths at the fastest speed possible.

  See Also:

   #WriteXY# #WriteXYCode#
}

Procedure WriteXYCode(const X,Y: Word;const Uts: String);
{
  Written: 96-11-16, Eddy L O Jansson (2:206/233)
  Updated: 96-11-23 (zerostring), 97-06-11 (multipage)

  Description:

   Write a string to X,Y. This version is special as the string may
   contain 'color-codes'. A change of attribute is encoded using the
   '~' symbol followed by the new attribute (byte/binary).

  Example:

   WriteXYCode(5,10,'~'+#$1B+'Norton~Novell~qRemote Access');

  See Also:

   #WriteXY# #WriteXYC#
}

Function  GetCharXY(const X,Y: Word): Byte;
{
  Written: 96-09-21, Eddy L O Jansson (2:206/233)
  Updated: 96-09-25, 97-06-11 (multipage)

  Description:

   Returns the character at location X,Y.

  See Also:

   #GetAttrXY# #GetCharAttrXY#
}

Function  GetAttrXY(const X,Y: Word): Byte;
{
  Written: 96-09-21, Eddy L O Jansson (2:206/233)
  Updated: 96-09-25, 97-06-11 (multipage)

  Description:

   Returns the attribute at location X,Y.

  See Also:

   #GetCharXY# #GetCharAttrXY#
}

Function  GetCharAttrXY(const X,Y: Word): Word;
{
  Written: 96-09-21, Eddy L O Jansson (2:206/233)
  Updated: 96-09-25, 97-06-11 (multipage)

  Description:

   Returns both the character and the attribute at location X,Y
   The attribute is returned in the hi-bits and the character in
   the low bits ($AACC).

  See Also:

  #SetCharAttrXY# #GetCharXY# #GetAttrXY# #SetCharXY# #SetAttrXY#
}

Procedure SetCharXY(const X,Y: Word;const Character: Byte);
{
  Written: 96-09-21, Eddy L O Jansson (2:206/233)
  Updated: 96-09-25, 97-06-11 (multipage)

  Description:

   Sets the character at location X,Y to Character.

  See Also:

   #GetAttrXY# #GetCharAttrXY# #SetAttrXY# #SetCharAttrXY#
}

Procedure SetAttrXY(const X,Y: Word;const Attribute: Byte);
{
  Written: 96-09-21, Eddy L O Jansson (2:206/233)
  Updated: 96-09-25, 97-06-11 (multipage)

  Description:

   Sets the attribute at location X,Y.

  See Also:

   #GetCharXY# #GetCharAttrXY# #SetCharXY# #SetCharAttrXY#
}

Procedure SetCharAttrXY(const X,Y,CharAttr: Word);
{
  Written: 96-09-21, Eddy L O Jansson (2:206/233)
  Updated: 96-09-25, 97-06-11 (multipage)

  Description:

   Sets both the character and the attribute at location X,Y
   The hi-bits contains the attribute and low bits the character ($AACC).

  See Also:

   #GetCharAttrXY# #GetCharXY# #GetAttrXY# #SetCharXY# #SetAttrXY#
}

Procedure Centre(const S :String;const Y: Word);
{
  Written: 96-09-21, Eddy L O Jansson (2:206/233)
  Updated: 96-09-25, 96-11-23 (zerostring), 97-06-11 (multipage)

  Description:

   Centre a string at line line Y.

  Notes:

   Does not change the corresponding attributes.

  See Also:

   #CentreC# #CentreCode#
}

Procedure CentreC(const S :String;const Y :Word;const Color :Byte);
{
  Written: 96-09-21, Eddy L O Jansson (2:206/233)
  Updated: 96-09-25, 96-11-23 (zerostring), 97-06-11 (multipage)

  Description:

   Centre a string written using attribute COLOR at line Y.

  See Also:

   #Centre# #CentreCode#
}

Procedure CentreCode(const S :String;const Y: Word);
{
  Written: 96-11-20, Eddy L O Jansson (2:206/233)
  Updated: 96-11-23 (zerostring), 97-06-11 (multipage)

  Description:

   Centre a string line Y. String may contain colorcodes (see #WriteXYCode#)

  See Also:

   #Centre# #CentreC#
}

Procedure BoxChar(const X,Y,Width,Height: Word;const Character :Byte);
{
  Written: 96-09-21, Eddy L O Jansson (2:206/233)
  Updated: 96-09-25, 97-06-11 (multipage)

  Description:

   Draw a BOX using Character (from X,Y to X+Width,Y+Height)

  Notes:

   Does not change the underlying attribute.

  See Also:

   #BoxAttr# #BoxCharAttr# #DrawH...# #DrawV...#
}

Procedure BoxAttr(const X,Y,Width,Height: Word;const Attribute :Byte);
{
  Written: 96-09-21, Eddy L O Jansson (2:206/233)
  Updated: 96-09-25, 97-06-11 (multipage)

  Description:

   Draw a BOX using Attribute (from X,Y to X+Width,Y+Height)

  Notes:

   Does not change the underlying characters.

  See Also:

   #BoxChar# #BoxCharAttr# #DrawH...# #DrawV...#
}

Procedure BoxCharAttr(const X,Y,Width,Height,CharAttr: Word);
{
  Written: 96-09-21, Eddy L O Jansson (2:206/233)
  Updated: 96-09-25, 97-06-11 (multipage)

  Description:

   Draw a BOX using Attribute & Character (from X,Y to X+Width,Y+Height)

  See Also:

   #BoxChar# #BoxAttr# #DrawH...# #DrawV...#
}

Procedure DrawHChar(const X,Y,Len: Word;const Character: Byte);
{
  Written: 96-09-21, Eddy L O Jansson (2:206/233)
  Updated: 96-09-25, 97-06-11 (multipage)

  Description:

   Set the character at X,Y to X+len,Y to a specific character.
   (ie it draws an horizontal bar/line )

  Notes:

   Does not change the underlying attribute.

  See Also:

   #DrawHAttr# #DrawHCharAttr# #DrawVChar# #DrawVAttr# #DrawVCharAttr#
}

Procedure DrawHAttr(const X,Y,Len: Word;const Attribute: Byte);
{
  Written: 96-09-21, Eddy L O Jansson (2:206/233)
  Updated: 96-09-25, 97-06-11 (multipage)

  Description:

   Set the attribute from X,Y to X+len,Y.
   (ie it draws an horizontal bar/line )

  Notes:

   Does not change the underlying characters.

  See Also

   #DrawHChar# #DrawHCharAttr# #DrawVChar# #DrawVAttr# #DrawVCharAttr#
}

Procedure DrawHCharAttr(const X,Y,Len,CharAttr: Word);
{
  Written: 96-09-25, Eddy L O Jansson (2:206/233)
  Updated: 97-06-11 (multipage)

  Description:

   Set the attribute and character from X,Y to X+len,Y.
   (ie it draws an horizontal bar/line )

  See Also

   #DrawHChar# #DrawHAttr# #DrawVChar# #DrawVAttr# #DrawVCharAttr#
}


Procedure DrawVChar(const X,Y,Height: Word;const Character: Byte);
{
  Written: 96-09-25, Eddy L O Jansson (2:206/233)
  Updated: 97-06-11 (multipage)

  Description:

   Draws a vertical line (or 'bar') using <character>.

  See also:

   #DrawVAttr# #DrawVCharAttr# #DrawHChar# #DrawHAttr# #DrawHCharAttr#

}
Procedure DrawVAttr(const X,Y,Height: Word;const Attribute: Byte);
{
  Written: 96-09-25, Eddy L O Jansson (2:206/233)
  Updated: 97-06-11 (multipage)

  Description:

   Draws a vertical line (or 'bar') using <attribute>.

  See also:

   #DrawVChar# #DrawVCharAttr# #DrawHChar# #DrawHAttr# #DrawHCharAttr#

}
Procedure DrawVCharAttr(const X,Y,Height,CharAttr: Word);
{
  Written: 96-09-25, Eddy L O Jansson (2:206/233)
  Updated: 97-06-11 (multipage)

  Description:

   Draws a vertical line (or 'bar') using character&attribute.

  See also:

   #DrawVChar# #DrawVAttr# #DrawHChar# #DrawHAttr# #DrawHCharAttr#

}

Procedure TextBoxGet(const X,Y,Width,Height: Word;const Destination: Pointer);
{
  Written: 96-11-16, Eddy L O Jansson (2:206/233)
  Updated: 97-04-18, Eddy (Example corrected), 97-06-11 (multipage)

  Description:

   Store the contents of the 'rectangle' X,Y to X+Width,Y+Height in
   <destination>. The base adress for the data is the adress stored
   in #TextMem#. Don't forget to reserve the memory before you utilize it.

  See Also:

   #TextBoxPut#

  Example:

   GetMem(ptrvar,(X*Y*2)+4);     (* X*Y words + 4 bytes for Width & Height *)
   TextBoxGet(X,Y,Width,Height,ptrvar); (* Copy ... *)
   TextBoxPut(X+Width,Y+Height,ptrvar); (* .. Paste *)
}

Procedure TextBoxPut(const X,Y: Word;const Source: Pointer);
{
  Written: 96-11-16, Eddy L O Jansson (2:206/233)
  Updated: 97-06-11 (multipage)

  Description:

   Restore the contents stored in pointer-variable <source> to
   location X,Y in/of #TextMem#. The format of the data-block is
    Width,Height    :Word
    Attr&Character  :Word; <repeated>

  See Also:

   #TextBoxGet#

}

Procedure Scroll(const X,Y,Width,Height,Direction,Attr: Byte);
{
  Written: 96-09-21, Eddy L O Jansson (2:206/233)
  Updated: 96-09-25

  Description:

   Scroll a portion of the screen up or down. X,Y,Width,Height defines
   the box to scroll. Direction is 0 for UP and 1 for DOWN. Attribute
   fills the blank line that is created.

  Notes:

   Uses BIOS. Will always scroll page 0.
}

Procedure SetBlink(const Mode :Boolean);
{
  Written: 96-09-25, Eddy L O Jansson (2:206/233)

  Description:

   Enable (TRUE) or Disable (FALSE) blink. Setting mode to FALSE will enable
   Highcolor (=bright) background colors.
}

Function Attr2F(const Attribute: Byte): Byte;
{
  Written: 97-01-10, Eddy L O Jansson (2:206/233)

  Description: Extract the foregroundcolor from an attribute.
  Returns a value between zero (black) and fifteen (bright white)
  If bit 4 (onebased) is set then the foregroundcolor is 'bright'.

  See Also:

  #Attr2B# #Attr2FB# #FB2Attr#
}

Function Attr2B(const Attribute: Byte): Byte;
{
  Written: 97-01-10, Eddy L O Jansson (2:206/233)

  Description: Extract the backgroundcolor from an attribute.
  Returns a value between zero (black) and seven (light gray/white)

  Notes:

  Important; Note how this function /won't/ extract the blink/intensity bit,
  something #Attr2FB# will.

  See Also:

  #Attr2F# #Attr2FB# #FB2Attr#
}

Procedure Attr2FB(const Attribute: Byte;var Foreground,Background: Byte);
{
  Written: 97-06-27, Eddy L O Jansson (2:206/233)

  Description: Extract the foreground and backgroundcolor from an attribute.

  Notes:

  Important; Note how this function /will/ extract the blink/intensity bit,
  something #Attr2B# will not.

  See Also:

  #Attr2F# #Attr2B# #FB2Attr#
}

Function FB2Attr(const Foreground,Background: Byte): Byte;
{
  Written: 97-01-10, Eddy L O Jansson (2:206/233)

  Description: Assemble foreground & background into an attribute byte.

  See Also:

  #Attr2F# #Attr2FB# #Attr2B#
}

Procedure WaitVBL;
 inline($BA/$DA/$03/$B4/$08/$EC/$84/$E0/$75/$FB/$EC/$84/$E0/$74/$FB);
 {
  Wait for vertical retrace.
 }

{$IFDEF MULTIPAGE}
Procedure PageCopy(const SourcePage,DestPage: Word);
{
  Written: 97-06-11, Eddy L O Jansson (2:206/233)

  Description:

   Copies the /visible/ part of a page to another.

   To copy the entire page instead, change the instruction;
    mov  cx,VisiblePageSize
   into
    mov  cx,PageSize

  Example:

   PageCopy(VDP,1); (* Copy "Visible Display Page" (active) to page 1 *)

}
{$ENDIF}

Procedure Move32(var Source,Destination;const Count :Word);
{
 Faster version of Move().
}

Procedure Font_Set8x8;

Procedure GotoXY(const X,Y: Byte);
{
  Written: 96-09-25, Eddy L O Jansson (2:206/233)
  Updated: 97-06-11 (multipage)

  Description:

   Position the cursor on the screen (in the WDP (Write Display Page)).

  Example:

   GotoXY(0,0);   (* Position cursor in upper left corner *)
   GotoXY(SW,SH); (* Position cursor in lower right corner *)
}

Procedure WhereXY(var X,Y: Byte);
{
  Written: 96-09-25, Eddy L O Jansson (2:206/233)

  Description:

   Return the X and Y position of the cursor on the screen (on the RDP
   (Read Display Page)). Returned coordinates are zerobased.

  See Also:

   #WhereX# #WhereY# #GotoXY#
}

Function  WhereX: Byte;
{
  Written: 96-09-25, Eddy L O Jansson (2:206/233)

  Description:

   Return the X position of the cursor on the screen (on the RDP
   (Read Display Page)). Returned coordinates are zerobased.

  See Also:

   #WhereXY# #WhereY# #GotoXY#
}

Function  WhereY: Byte;
{
  Written: 96-09-25, Eddy L O Jansson (2:206/233)

  Description:

   Return the Y position of the cursor on the screen (on the RDP
   (Read Display Page)). Returned coordinates are zerobased.

  See Also:

   #WhereXY# #WhereX# #GotoXY#
}

Function  GetCursor: Word;
{
 Description:

 Return the current cursor shape.
}

Procedure SetCursor(const Cursor: Word);
{
 Description:

 Set the current cursor shape.
}

Procedure HideCursor;
{
 Description:

 Remove the cursor from screen by modifying its baseline & height
}

Procedure NormalCursor;
{
 Description:

 Sets a "normal" cursorshape ("insert mode")}

Procedure BlockCursor;
{
 Description:

 Set the cursorshape into a block ("overwrite mode")
}

{-[ Implementation / Code ]-------------------------------------------------}

IMPLEMENTATION

Procedure SetScreen(const mode: byte); Assembler;
asm
 mov ah,0
 mov al,mode
 mov CSM,al
 int $10
end;

{$IFDEF MULTIPAGE}
Procedure SetDisplayPage(const page: word); Assembler;
asm
 mov ax,page
 mov VDP,ax
 mov ah,5
 int $10
end;
{$ENDIF}

Function SetVESAMode(const mode: Word): Byte; Assembler;
asm
 mov ax,4F02h
 mov bx,mode
 int 10h
end;

Function GetScreen: Byte; Assembler;
asm
 mov  ah,15
 int  $10
end;

Function GetScreenWidth: Byte; Assembler;
asm
 mov  ah,15
 int  $10
 mov  al,ah
end;

Function GetScreenHeight: Byte; Assembler;
asm
 mov  ax,$40
 mov  es,ax
 mov  al,byte ptr es:[$84]
 inc  al
end;

Procedure ClrScr; Assembler;
asm
 les  di,TextMem
 {$IFDEF MULTIPAGE}
 mov  ax,WDP
 mov  bx,PageSize
 mul  bx
 add  di,ax
 {$ENDIF}
 mov  cx,VisiblePageSize
 shr  cx,1
 mov  ax,$0720      { Attribute 7, character 32 (space) }
 rep  stosw
end;

Procedure FillScreen(const AttrChar :Word); Assembler;
asm
  les  di,TextMem
  {$IFDEF MULTIPAGE}
  mov  ax,WDP
  mov  bx,PageSize
  mul  bx
  add  di,ax
 {$ENDIF}
  mov  cx,VisiblePageSize
  shr  cx,1
  mov  ax,AttrChar
  cld
  rep  stosw
end;

Procedure WriteXY(const X,Y: Word;const Uts: String); Assembler;
asm
 push ds
 les  di,TextMem
 {$IFDEF MULTIPAGE}
 mov  ax,WDP
 mov  bx,PageSize
 mul  bx
 add  di,ax
 {$ENDIF}
 mov  bx,Y
 add  bx,bx
 add  di,word ptr CYTable[bx]
 mov  bx,X
 add  bx,bx
 add  di,bx
 lds  si,UTS
 xor  ch,ch
 mov  cl,byte ptr[si]
 jcxz @exit
 inc  si
 cld
@next:
 movsb
 inc  di
 loop @next
@exit:
 pop  ds
end;

Procedure WriteXYC(const X,Y: Word;const Color: Byte;const Uts: String); Assembler;
asm
 push ds
 les  di,TextMem
 {$IFDEF MULTIPAGE}
 mov  ax,WDP
 mov  bx,PageSize
 mul  bx
 add  di,ax
 {$ENDIF}
 mov  bx,Y
 add  bx,bx
 add  di,word ptr CYTable[bx]
 mov  bx,X
 add  bx,bx
 add  di,bx
 lds  si,UTS
 xor  ch,ch
 mov  cl,byte ptr[si]
 jcxz @exit
 inc  si
 mov  ah,COLOR
 cld
@next:
 lodsb          { get character inte AL }
 stosw          { write character + color }
 loop @next
@exit:
 pop  ds
end;

Procedure WriteXYCode(const X,Y: Word;const Uts: String); Assembler;
asm
 push ds
 les  di,TextMem
 {$IFDEF MULTIPAGE}
 mov  ax,WDP
 mov  bx,PageSize
 mul  bx
 add  di,ax
 {$ENDIF}
 mov  bx,Y
 add  bx,bx
 add  di,word ptr CYTable[bx]
 mov  bx,X
 add  bx,bx
 add  di,bx
 lds  si,UTS
 xor  ch,ch
 mov  cl,byte ptr[si]
 jcxz @exit
 inc  si
 cld
@next:
 lodsb          { get character inte AL }
 cmp  al,'~'
 je   @getcolor
 stosw          { write character + color }
 loop @next
 jmp  @exit
@getcolor:
 lodsb
 sub  cl,2
 mov  ah,al
 jmp  @next
@exit:
 pop  ds
end;

Function GetCharXY(const X,Y: Word): Byte; Assembler;
asm
 les  di,TextMem
 {$IFDEF MULTIPAGE}
 mov  ax,RDP
 mov  bx,PageSize
 mul  bx
 add  di,ax
 {$ENDIF}
 mov  bx,Y
 add  bx,bx
 add  di,word ptr CYTable[bx]
 mov  bx,X
 add  bx,bx
 add  di,bx
 mov  al,byte ptr es:[di]
end;

Function GetAttrXY(const X,Y: Word): Byte; Assembler;
asm
 les  di,TextMem
 {$IFDEF MULTIPAGE}
 mov  ax,RDP
 mov  bx,PageSize
 mul  bx
 add  di,ax
 {$ENDIF}
 mov  bx,Y
 add  bx,bx
 add  di,word ptr CYTable[bx]
 mov  bx,X
 add  bx,bx
 add  di,bx
 mov  al,byte ptr es:[di+1]
end;

Function GetCharAttrXY(const X,Y: Word): Word; Assembler;
asm
 les  di,TextMem
 {$IFDEF MULTIPAGE}
 mov  ax,RDP
 mov  bx,PageSize
 mul  bx
 add  di,ax
 {$ENDIF}
 mov  bx,Y
 add  bx,bx
 add  di,word ptr CYTable[bx]
 mov  bx,X
 add  bx,bx
 add  di,bx
 mov  ax,word ptr es:[di]
end;

Procedure SetCharXY(const X,Y: Word;const Character: Byte); Assembler;
asm
 les  di,TextMem
 {$IFDEF MULTIPAGE}
 mov  ax,WDP
 mov  bx,PageSize
 mul  bx
 add  di,ax
 {$ENDIF}
 mov  bx,Y
 add  bx,bx
 add  di,word ptr CYTable[bx]
 mov  bx,X
 add  bx,bx
 add  di,bx
 mov  al,CHARACTER
 mov  byte ptr es:[di],al
end;

Procedure SetAttrXY(const X,Y: Word;const Attribute: Byte); Assembler;
asm
 les  di,TextMem
 {$IFDEF MULTIPAGE}
 mov  ax,WDP
 mov  bx,PageSize
 mul  bx
 add  di,ax
 {$ENDIF}
 mov  bx,Y
 add  bx,bx
 add  di,word ptr CYTable[bx]
 mov  bx,X
 add  bx,bx
 add  di,bx
 mov  al,ATTRIBUTE
 mov  byte ptr es:[di+1],al
end;

Procedure SetCharAttrXY(const X,Y,CharAttr: Word); Assembler;
asm
 les  di,TextMem
 {$IFDEF MULTIPAGE}
 mov  ax,WDP
 mov  bx,PageSize
 mul  bx
 add  di,ax
 {$ENDIF}
 mov  bx,Y
 add  bx,bx
 add  di,word ptr CYTable[bx]
 mov  bx,X
 add  bx,bx
 add  di,bx
 mov  ax,CHARATTR
 mov  word ptr es:[di],ax
end;

Procedure Centre(const S :String;const Y: Word); Assembler;
asm
 push ds
 les  di,TextMem
 {$IFDEF MULTIPAGE}
 mov  ax,WDP
 mov  bx,PageSize
 mul  bx
 add  di,ax
 {$ENDIF}
 mov  bx,Y
 add  bx,bx
 add  di,word ptr CYTable[bx] { di = correct line }
 mov  bx,SW
 xor  ah,ah
 lds  si,S
 mov  al,byte ptr [si]        { get string length }
 cmp  al,0
 je   @exit
 mov  cl,al                   { save length }
 sub  bx,ax                   { Screenwidth-stringlength }
 and  bx,$FE                  { make even by removing LSB (better to set?) }
 add  di,bx                   { calc final destination }
 xor  ch,ch
 inc  si
 cld
@next:
 movsb
 inc  di
 loop @next
@exit:
 pop  ds
end;

Procedure CentreC(const S :String;const Y :Word;const Color :Byte); Assembler;
asm
 push ds
 les  di,TextMem
 {$IFDEF MULTIPAGE}
 mov  ax,WDP
 mov  bx,PageSize
 mul  bx
 add  di,ax
 {$ENDIF}
 mov  bx,Y
 add  bx,bx
 add  di,word ptr CYTable[bx] { di = correct line }
 mov  bx,SW
 xor  ah,ah
 lds  si,S
 mov  al,byte ptr [si]        { get string length }
 cmp  al,0
 je   @exit
 mov  cl,al                   { save length }
 sub  bx,ax                   { Screenwidth-stringlength }
 and  bx,$FE                  { make even by removing LSB (better to set?) }
 add  di,bx                   { calc final destination }
 xor  ch,ch
 inc  si
 mov  ah,COLOR
 cld
@next:
 lodsb          { get character inte AL }
 stosw          { write character + color }
 loop @next
@exit:
 pop  ds
end;

Procedure CentreCode(const S :String;const Y: Word); Assembler;
asm
 push ds
 les  di,TextMem
 {$IFDEF MULTIPAGE}
 mov  ax,WDP
 mov  bx,PageSize
 mul  bx
 add  di,ax
 {$ENDIF}
 mov  bx,Y
 add  bx,bx
 add  di,word ptr CYTable[bx] { di = correct line }
 mov  bx,SW
 xor  ah,ah
 lds  si,S
 mov  dx,si
 mov  al,byte ptr [si]        { get string length }
 cmp  al,0
 je   @exit
 mov  cl,al                   { save length }
 sub  bx,ax                   { Screenwidth-stringlength }
 and  bx,$FE                  { make even by removing LSB (better to set?) }
 add  di,bx                   { calc final destination }
 xor  ch,ch
 mov  ah,al
@nextc:
 dec  cl
 jz   @proceed
 lodsb
 cmp  al,'~'
 jne  @nextc
 inc  ch
 jmp  @nextc
@proceed:
 mov si,dx
 mov cl,ch
 shl cl,1
 xor ch,ch
 add di,cx
 mov cl,ah
 inc si
 cld
@next:
 lodsb          { get character inte AL }
 cmp  al,'~'
 je   @getcolor
 stosw          { write character + color }
 loop @next
 jmp  @exit
@getcolor:
 lodsb
 sub  cl,2
 mov  ah,al
 jmp  @next
@exit:
 pop  ds
end;

Procedure BoxChar(const X,Y,Width,Height: Word;const Character :Byte); Assembler;
{ ** OPTIMIZE THIS! }
asm
 push ds
 xor  cx,cx
@nextline:
 les  di,TextMem                 { ugh, not nice :-( }
 {$IFDEF MULTIPAGE}
 mov  ax,WDP
 mov  bx,PageSize
 mul  bx
 add  di,ax
 {$ENDIF}
 mov  bx,Y
 add  bx,cx                      { Add HEIGHT counter }
 add  bx,bx
 add  di,word ptr CYTable[bx]
 mov  bx,X
 add  bx,bx
 add  di,bx

 mov  bx,cx
 mov  cx,WIDTH
 mov  al,CHARACTER
 cld
@next:
 stosb
 inc  di
 loop @next
 mov  cx,bx
 inc  cx
 cmp  cx,HEIGHT
 jne  @nextline
@exit:
 pop  ds
end;

Procedure BoxAttr(const X,Y,Width,Height: Word;const Attribute :Byte); Assembler;
{ ** OPTIMIZE THIS  }
asm
 push ds
 xor  cx,cx
@nextline:
 les  di,TextMem
 {$IFDEF MULTIPAGE}
 mov  ax,WDP
 mov  bx,PageSize
 mul  bx
 add  di,ax
 {$ENDIF}
 mov  bx,Y
 add  bx,cx                      { Add HEIGHT counter }
 add  bx,bx
 add  di,word ptr CYTable[bx]
 mov  bx,X
 add  bx,bx
 add  di,bx

 mov  bx,cx
 mov  cx,WIDTH
 mov  al,ATTRIBUTE
 cld
@next:
 inc  di
 stosb
 loop @next
 mov  cx,bx
 inc  cx
 cmp  cx,HEIGHT
 jne  @nextline
@exit:
 pop  ds
end;

Procedure BoxCharAttr(const X,Y,Width,Height,CharAttr: Word); Assembler;
asm
 push ds
 xor  cx,cx
@nextline:
 les  di,TextMem
 {$IFDEF MULTIPAGE}
 mov  ax,WDP
 mov  bx,PageSize
 mul  bx
 add  di,ax
 {$ENDIF}
 mov  bx,Y
 add  bx,cx
 add  bx,bx
 add  di,word ptr CYTable[bx]
 mov  bx,X
 add  bx,bx
 add  di,bx

 mov  bx,cx
 mov  cx,WIDTH
 mov  ax,CHARATTR
 cld
@next:
 stosw
 loop @next
 mov  cx,bx
 inc  cx
 cmp  cx,HEIGHT
 jne  @nextline
@exit:
 pop  ds
end;

Procedure DrawHChar(const X,Y,Len: Word;const Character: Byte); Assembler;
asm
 push ds
 les  di,TextMem
 {$IFDEF MULTIPAGE}
 mov  ax,WDP
 mov  bx,PageSize
 mul  bx
 add  di,ax
 {$ENDIF}
 mov  bx,Y
 add  bx,bx
 add  di,word ptr CYTable[bx]
 mov  bx,X
 add  bx,bx
 add  di,bx
 mov  cx,LEN
 mov  al,CHARACTER
 cld
@next:
 stosb
 inc  di
 loop @next
 pop  ds
end;

Procedure DrawHAttr(const X,Y,Len: Word;const Attribute: Byte); Assembler;
asm
 push ds
 les  di,TextMem
 {$IFDEF MULTIPAGE}
 mov  ax,WDP
 mov  bx,PageSize
 mul  bx
 add  di,ax
 {$ENDIF}
 mov  bx,Y
 add  bx,bx
 add  di,word ptr CYTable[bx]
 mov  bx,X
 add  bx,bx
 add  di,bx
 mov  cx,LEN
 mov  al,ATTRIBUTE
 cld
@next:
 inc  di
 stosb
 loop @next
 pop  ds
end;

Procedure DrawHCharAttr(const X,Y,Len,CharAttr: Word); Assembler;
asm
 push ds
 les  di,TextMem
 {$IFDEF MULTIPAGE}
 mov  ax,WDP
 mov  bx,PageSize
 mul  bx
 add  di,ax
 {$ENDIF}
 mov  bx,Y
 add  bx,bx
 add  di,word ptr CYTable[bx]
 mov  bx,X
 add  bx,bx
 add  di,bx
 mov  cx,LEN
 mov  ax,CHARATTR
 cld
@next:
 stosw
 loop @next
 pop  ds
end;

Procedure DrawVChar(const X,Y,Height: Word;const Character: Byte); Assembler;
asm
 les  di,TextMem
 {$IFDEF MULTIPAGE}
 mov  ax,WDP
 mov  bx,PageSize
 mul  bx
 add  di,ax
 {$ENDIF}
 mov  bx,Y
 add  bx,bx
 add  di,word ptr CYTable[bx]
 mov  bx,X
 add  bx,bx
 add  di,bx
 mov  al,CHARACTER
 mov  cx,HEIGHT
 mov  bx,word ptr CYtable[2]
@nextline:
 mov  byte ptr es:[di],al
 add  di,bx
 loop @nextline
end;

Procedure DrawVAttr(const X,Y,Height: Word;const Attribute: Byte); Assembler;
asm
 les  di,TextMem
 {$IFDEF MULTIPAGE}
 mov  ax,WDP
 mov  bx,PageSize
 mul  bx
 add  di,ax
 {$ENDIF}
 mov  bx,Y
 add  bx,bx
 add  di,word ptr CYTable[bx]
 mov  bx,X
 add  bx,bx
 add  di,bx
 mov  al,ATTRIBUTE
 mov  cx,HEIGHT
 mov  bx,word ptr CYtable[2]
@nextline:
 mov  byte ptr es:[di+1],al
 add  di,bx
 loop @nextline
end;

Procedure DrawVCharAttr(const X,Y,Height,CharAttr: Word); Assembler;
asm
 les  di,TextMem
 {$IFDEF MULTIPAGE}
 mov  ax,WDP
 mov  bx,PageSize
 mul  bx
 add  di,ax
 {$ENDIF}
 mov  bx,Y
 add  bx,bx
 add  di,word ptr CYTable[bx]
 mov  bx,X
 add  bx,bx
 add  di,bx
 mov  ax,CHARATTR
 mov  cx,HEIGHT
 mov  bx,word ptr CYtable[2]
@nextline:
 mov  word ptr es:[di],ax
 add  di,bx
 loop @nextline
end;

Procedure TextBoxGet(const X,Y,Width,Height: Word;const Destination: Pointer); Assembler;
asm
 push ds
 {$IFDEF MULTIPAGE}
 mov  ax,RDP
 mov  bx,PageSize
 mul  bx
 mov  si,ax
 {$ENDIF}
 mov  bx,Y
 add  bx,bx
 mov  dx,word ptr CYtable[2]  { get screen width in bytes }
 add  si,word ptr CYTable[bx] { get start addr for this Y-coord }
 mov  bx,X                    { get X-coord }
 add  bx,bx                   { X-coord:=Xcoord * 2 ((attr+char)=2 bytes) }
 add  bx,si                   { store source X-offset in BX }
 lds  si,TextMEM              { setup DS:SI as source }
 add  si,bx                   { add (width*2)*height to SI }
                              { DS:SI now points to first byte to copy }
 les  di,Destination          { ES:DI points to destination }
 mov  ax,width                { Get width }
 cld
 stosw                        { Store width of block at ES:DI ; inc DI,2 }
 mov  bx,ax                   { Save width in BX for future use. }
 shl  ax,1
 sub  dx,ax                   { DX = precalcued 'add-value' for start }
                              {      of the next line }
 mov  ax,height
 stosw                        { Store height of block at ES:DI ; inc DI,2 }
@nextline:
 mov  cx,bx                   { Get width }
 rep  movsw
 add  si,dx                   { Go to start of next line }
 dec  ax                      { decrease "height" }
 jnz  @nextline
@exit:
 pop ds
end;

Procedure TextBoxPut(const X,Y: Word;const Source: Pointer); Assembler;
{
 **REMOVE/OBSOLETE:

 Benchmarking: (on an Intel-P133 with CL-GD5436(PCI) Graphics-card)

  TextBoxPut() will draw 5,676 blocks a 60x20 characters (=6,811,200 bytes)
   in  4 ticks (Virtual-buffer) (0.22s)
   in 25 ticks (Videomem)       (1.38s)
}
asm
 push ds
 mov  dx,word ptr CYtable[2]
 mov  bx,Y
 add  bx,bx
 mov  si,word ptr CYTable[bx]
 mov  bx,X
 add  bx,bx
 add  bx,si
 les  di,TextMEM
 add  di,bx
 {$IFDEF MULTIPAGE}
 push dx
 mov  ax,WDP
 mov  bx,PageSize
 mul  bx
 add  di,ax
 pop  dx
 {$ENDIF}
 lds  si,Source
 cld
 lodsw           { Get width into AX }
 mov  bx,ax
 shl  ax,1
 sub  dx,ax
 lodsw           { get height into AX }
@nextline:
 mov  cx,bx
 rep  movsw
 add  di,dx
 dec  ax
 jnz  @nextline
@exit:
 pop ds
end;

Procedure Scroll(const X,Y,Width,Height,Direction,Attr: Byte); Assembler;
{ 0 = Up, 1 = Down }
asm
 mov bh,ATTR
 mov ch,Y
 mov cl,X
 mov dh,ch
 add dh,HEIGHT
 dec dh
 mov dl,cl
 add dl,WIDTH
 dec dl
 mov al,1                      { Lines to scroll }
 mov ah,DIRECTION
 and ah,1
 or  ah,6
 int $10
end;

Procedure SetBlink(const Mode :Boolean); Assembler;
asm
 mov  ax,$1003
 mov  bl,mode
 int  $10
end;

{
  76543210 AL
          blue component of foreground color
         green component of foreground color
        red component of foreground color
       INTENSITY component of foreground color
      blue component of background color
     green component of background color
    red component of background color
   BLINKING of foreground character
}
{** Inline'ize these three functions }
Function Attr2B(const Attribute: Byte): Byte; Assembler;
asm
 mov al,attribute
 shr al,4
 and al,1+2+4
end;

Function Attr2F(const Attribute: Byte): Byte; Assembler;
asm
 mov al,attribute
 and al,1+2+4+8
end;

Procedure Attr2FB(const Attribute: Byte;var Foreground,Background: Byte); Assembler;
asm
 mov  al,attribute
 mov  ah,al
 les  di,Foreground
 and  al,1+2+4+8
 mov  byte ptr es:[di],al
 shr  ah,4
 and  ah,1+2+4+8 { -8}
 mov  byte ptr es:[di+1],ah
end;

Function FB2Attr(const Foreground,Background: Byte): Byte; Assembler;
asm
 mov al,background
 shl al,4
 or  al,foreground
end;

Function GetCursor: Word; Assembler;
asm
 mov ah,3
 int $10
 mov ax,cx
end;

Procedure SetCursor(const Cursor: Word); Assembler;
asm
 mov ah,1
 mov cx,Cursor
 int $10
end;

Procedure HideCursor; Assembler;
asm
 mov ah,1
 mov cx,$2000 { some say this ain't the way to do it. I don't care >:-) }
 int $10
end;

Procedure NormalCursor; Assembler;
asm
 mov ah,1
 mov ch,6
 mov cl,7
 int $10
end;

Procedure BlockCursor; Assembler;
asm
 mov ah,1
 mov ch,0
 mov cl,7
 int $10
end;

Procedure GotoXY(const X,Y: Byte); Assembler;
asm
 {$IFDEF MULTIPAGE}
 mov  bx,WDP
 shl  bx,8
 {$ELSE}
 xor  bh,bh
 {$ENDIF}
 mov  ah,2
 mov  dh,Y
 mov  dl,X
 int  $10
end;

Procedure WhereXY(var X,Y: Byte); Assembler;
asm
 mov  ah,3
 {$IFDEF MULTIPAGE}
 mov  bx,RDP
 shl  bx,8
 {$ELSE}
 xor  bh,bh
 {$ENDIF}
 int  $10
 les  di,X
 mov  byte ptr es:[di],dl
 mov  byte ptr es:[di+1],dh
end;

Function  WhereX: Byte; Assembler;
asm
 mov  ah,3
 {$IFDEF MULTIPAGE}
 mov  bx,RDP
 shl  bx,8
 {$ELSE}
 xor  bh,bh
 {$ENDIF}
 int  $10
 mov  al,dl
end;

Function  WhereY: Byte; Assembler;
asm
 mov  ah,3
 {$IFDEF MULTIPAGE}
 mov  bx,RDP
 shl  bx,8
 {$ELSE}
 xor  bh,bh
 {$ENDIF}
 int  $10
 mov  al,dh
end;

{$IFDEF MULTIPAGE}
Procedure PageCopy(const SourcePage,DestPage: Word); Assembler;
asm
 push ds
 les  di,TextMem
 mov  ax,DestPage
 mov  bx,PageSize
 mul  bx
 add  di,ax
 mov  ax,SourcePage
 mov  bx,PageSize
 mov  cx,VisiblePageSize
 shr  cx,1
 lds  si,TextMem
 mul  bx
 add  si,ax
 cld
 rep  movsw
 pop  ds
end;
{$ENDIF}

Procedure Move32(var Source,Destination;const Count :Word); Assembler;
asm
 cld
 mov  dx,ds
 les  di,destination
 lds  si,source
 mov  cx,count
 shr  cx,1
 jnc  @mw
 movsb
@mw:
 shr  cx,1
 jnc  @md
 movsw
@md:
 db   66h; rep movsw
 mov  ds,dx
end;

Procedure Font_Set8x8; Assembler;
asm
 mov ax,1112h
 xor bl,bl
 int $10
end;

Procedure Init;
begin
 if (Mem[$0000:$0449] in [$7..$20]) then
  TextMem:=Ptr($B000,0) else TextMem:=Ptr($B800,0);
 asm
  { Setup "SW", "SH" & "PreviousVideoMode" }
  mov  ah,15
  int  $10

  {**TO DO: Setup VDP,RDP and WDP! }
  {$IFDEF MULTIPAGE }
  mov  VDP,0
  mov  RDP,0
  mov  WDP,0
  {$ENDIF}
  mov  CSM,al
  mov  PreviousVideoMode,al
  mov  al,ah
  xor  ah,ah
  mov  SW,ax
  mov  ax,$0040
  mov  es,ax
  mov  al,byte ptr es:[$84]
  inc  al
  mov  SH,ax

  mov  ax,word ptr es:[$4C]
  mov  PageSize,ax

  { Setup "CYTable" }
  mov di,offset CYTable
  mov ax,ds
  mov es,ax
  mov cx,MaxScreenHeight
  mov bx,SW
  shl bx,1
  add di,cx
  add di,cx
  std
 @calcloop:
  mov  ax,cx
  mul  bx
  stosw
  loop @calcloop
  cld
  { Setup "VisiblePageSize" }
  mov  ax,SH
  mul  bx
  mov  VisiblePageSize,ax
 end;

end;

BEGIN
{ Setup/Initialization }
{ STMR_LOW.Init; }
END.
