Unit STMR_STR; { Sarumans Text Mode Routines (String Manipulation) }
{
  STMR_STR v0.0.1 - (C)1996,97 Eddy L O Jansson / DFR Research & Engineering


  TO DO:

    Certify all routines;
     * Work according to specs (outdata is the wanted product of indata)
     * Work within bounds (min-max) (don't destroy any other vars/data)
     * Work with empty strings
     * Be as fast as possible
     * Be as dynamic as possible
}

{-[ Function Interface ]----------------------------------------------------}
INTERFACE

type
 ByteArray      =Array[0..255] of byte;

const
 XLT_Swedish :ByteArray = ( { Upper-table /w support for Swedish chars }
  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15, 16, 17, 18,
 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37,
 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56,
 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75,
 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94,
 95, 96, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81,
 82, 83, 84, 85, 86, 87, 88, 89, 90,123,124,125,126,127,128,129,130,131,142,
133,143,135,136,137,138,139,140,141,142,143,144,145,146,147,153,149,150,151,
152,153,154,155,156,157,158,159,160,161,162,163,164,165,166,167,168,169,170,
171,172,173,174,175,176,177,178,179,180,181,182,183,184,185,186,187,188,189,
190,191,192,193,194,195,196,197,198,199,200,201,202,203,204,205,206,207,208,
209,210,211,212,213,214,215,216,217,218,219,220,221,222,223,224,225,226,227,
228,229,230,231,232,233,234,235,236,237,238,239,240,241,242,243,244,245,246,
247,248,249,250,251,252,253,254,255);


Function Left(const S: String;const n: Byte): String;
 { Returns the first <nr> characters to the left of the string <s> }

Function Right(const S: String;const n: Byte): String;
 { Returns the first <nr> characters to the right of the string <s> }

Function Copy(const Source: String;const Start,Count: Byte): String;
 { Replacement for the dos/system 'Copy' function }

Procedure Copy_P(const Source: string;const Start,Count: Byte;var Destination: String);
 { As 'copy' but in procedural form and ofcourse faster (much faster) }

Procedure DOS_ReadLn(var S;const MaxLen: Byte);
 { Buffered string input from a standard console device. }

Procedure XLAT(var Source: String; var Table);
 { Translate a string (buffer) using an translation-table and XLAT }

Procedure Upper_P(var S: String);
 { Uppercase using DOS services. Uses country-specific translationtable }

Function Upper_F(const S: String): String;
 { Uppercase using DOS services. Uses country-specific translationtable }

Procedure Strip(var S;const StripChar: Char);
 { Remove any leading and trailing <stripchars> from a string }

Function RepeatChar(const C: Char;const N: Word): String;
 { Create a string consisting of N number of <C>hars }

Function Instr(const Source,Sub: String;const StartPos: Byte): Byte;
 { Returns the position of the first occurence of the string <sub>
   in the string <source>, starting at byte <startpos>

   Pos() is faster unless you can benefit from the <startpos> parameter.
  }

Procedure PadRight_P(var S: String;const Count,PadChar: Byte);
 { Pad a string at the right side. Characters are /added/ to the right
   side until the string is <count> characters long. if Count is less
   then the string-length the string will be truncated to <count> chars.
 }

Function PadRight_F(const S: String;const Count,PadChar: Byte): String;
 { Pad a string at the right side. Characters are /added/ to the right
   side until the string is <count> characters long. if Count is less
   then the string-length the string will be truncated to <count> chars.
 }

{ Function PadLeft_x() }

Function StripFromRight(const S: String;const BreakChar: Char): String;
 { Search a string from right to left, and return the part left of the
   first occurence of <breakchar> found. If no <breakchar> is found then
   the whole string is returned }

{ StripFromLeft() }

Procedure AddChar(Var Str: String;C :Char);
 { Add a character to a string. Faster than S1:=S2+char }
Inline($58/$5F/$07/$26/$FE/$05/$31/$DB/$26/$8A/$1D/$01/$DF/$AA);

{-[ Implementation / Code ]-------------------------------------------------}
IMPLEMENTATION

Function Left(const S: String;const n: Byte): String; Assembler;
{ Updated: 97-04-16 Eddy L O Jansson (routine checked/certified) }
asm
 mov  dx,ds
 cld
 les  di,@result
 lds  si,S
 lodsb
 xor  ch,ch
 mov  cl,n
 cmp  cl,al
 jbe  @doit
 mov  cl,al
@doit:
 mov  byte ptr es:[di],cl
 inc  di
 rep  movsb
 mov  ds,dx
end;

Function Right(const S: String;const n: Byte): String; Assembler;
{ Updated: 97-04-16 Eddy L O Jansson (routine checked/certified)
    Notes: DF will be set upon return from this routine }
asm
 mov  dx,ds
 les  di,@result
 lds  si,S
 xor  ch,ch
 xor  ah,ah
 mov  cl,n
 mov  al,byte ptr [si]
 cmp  cl,al
 jbe  @doit
 mov  cl,al
@doit:
 mov  byte ptr es:[di],cl
 add  si,ax
 add  di,cx
 std
 rep  movsb
 mov  ds,dx
end;

Function Copy(const Source: String;const Start,Count: Byte): String; Assembler;
asm
 push ds
 les  di,@result
 lds  si,source    { load in DS:SI pointer to str1 }
 mov  ah,[si]
 and  ah,ah
 je   @null        { was a zero-string? }
 mov  bl,start
 cmp  ah,bl        { start > length of source? }
 jb   @null        { yes, empty output string }
 mov  al,count
 mov  dh,al
 add  dh,bl
 dec  dh
 cmp  ah,dh        { start + nrchars - 1 > length of source? }
 jb   @rest        { yes, copy rest of source }
 jmp  @copy
@null:
 mov  al,0
 jmp  @done
@rest:
 sub  ah,bl
 inc  ah
 mov  al,ah        { length of source - start }
@copy:
 mov  cl,al
 xor  ch,ch
 xor  bh,bh
 add  si,bx
 mov  dx,di        { store pointer to output string }
 inc  di
 cld
 rep  movsb        { copy part of source to dest }
 mov  di,dx
@done:
 mov  es:[di],al   { set length of destination }
 pop  ds
end;

Procedure Copy_P(const Source: string;const Start,Count: Byte;var Destination: String); Assembler;
asm
 push ds
 les  di,destination
 lds  si,source    { load in DS:SI pointer to str1 }
 mov  ah,[si]
 and  ah,ah
 je   @null        { was a zero-string? }
 mov  bl,start
 cmp  ah,bl        { start > length of source? }
 jb   @null        { yes, empty output string }
 mov  al,count
 mov  dh,al
 add  dh,bl
 dec  dh
 cmp  ah,dh        { start + nrchars - 1 > length of source? }
 jb   @rest        { yes, copy rest of source }
 jmp  @copy
@null:
 mov  al,0
 jmp  @done
@rest:
 sub  ah,bl
 inc  ah
 mov  al,ah        { length of source - start }
@copy:
 mov  cl,al
 xor  ch,ch
 xor  bh,bh
 add  si,bx
 mov  dx,di        { store pointer to output string }
 inc  di
 cld
 rep  movsb        { copy part of source to dest }
 mov  di,dx
@done:
 mov  [di],al      { set length of destination }
 pop  ds
end;


Procedure DOS_ReadLn(var S;const MaxLen: Byte); Assembler;
{ Buffered string input from a standard console device }
asm
 cld
 push ds
 lds  si,S
 mov  dx,si        { DS:DX = Buffer }
 mov  ah,10
 mov  bl,MaxLen
 inc  bl
 mov  [si],bl      { Load buffer with MaxLen }
 int  $21
 les  di,S         { Must move string back two bytes, over buffer }
 inc  si
 lodsb
 stosb
 xor  ch,ch
 mov  cl,al
 rep  movsb
 pop  ds
end;

Function Upper_F(const S: String): String; Assembler;
asm
 push ds
 lds  si,s
 cld
 lodsb
 mov  cl,al
 xor  ch,ch
 jcxz @exit
 les  di,@result
 stosb
 mov  dx,di
 push cx
 rep  movsb
 push es
 pop  ds
 pop  cx
 mov  ax,$6521
 int  $21
@exit:
 pop  ds
end;

Procedure Upper_P(var S: String); Assembler;
asm
 push ds
 lds  si,s
 mov  cl,[si]
 xor  ch,ch
 jcxz @exit
 lea  dx,[si+1]
 mov  ax,$6521
 int  $21
@exit:
 pop  ds
end;

Procedure XLAT(var Source: String; var Table); Assembler;
asm
 mov  dx,ds        { save ds }
 lds  bx,Table     { load ds:bx with Table address }
 les  di,Source    { load es:di with Source address }
 seges             { override ds segment}
 mov  al,[di]      { load al with length of source }
 xor  ah,ah        { set ah to zero, we need a word for cx }
 mov  cx,ax        { assign length of source to counter }
 jcxz @end         { if cx = 0 exit}
 inc  di           { increment di & skip length byte on 1st pass }
@filter:
 mov  al,[di]      { load byte in ax from es:di }
 xlat              { translate... }
 mov  [di],al      { send byte to es:di }
 inc  di           { increment di }
 loop @filter      { decrement cx and loop back if cx > 0 }
@end:
 mov  ds,dx        { restore ds }
end;

Procedure Strip(var S;const StripChar: Char); Assembler;
asm
  push ds
  les  di,S
  mov  ax,es
  mov  ds,ax
  mov  bx,di        { save startpos }
  mov  al,StripChar
  xor  ch,ch
  mov  cl,[di]      { length in CL }
  jcxz @quit        { exit if null }
 {-- scan tail --}
  add  di,cx        { start at far end of string }
  cmp  al,[di]      { check last character }
  jne  @scanfront   { exit if no match }
  std
  repe scasb        { search for first non-matching byte }
  je   @null        { if you are stripping all characters }
  inc  cl           { fix length }
  mov  [bx],cl      { new length }
@scanfront:
  mov  di,bx
  inc  di
  cmp  al,[di]      { check first char }
  jne  @quit        { exit if no match }
  inc  di           { next char }
  dec  cl
  cld               { frontwards }
  repe scasb        { do search after first }
  je   @null
  mov  si,di        { char after last match }
  mov  di,bx        { to start of string }
  inc  cx
  inc  di
  dec  si
@null:
  mov  [bx],cl
  rep  movsb
@quit:
  pop  ds
end;

Function RepeatChar(const C: Char;const N: Word): String; Assembler;
asm
 mov  cx,N
 mov  al,C
 les  di,@result
 mov  byte ptr es:[di],cl
 inc  di
 cld
 rep  stosb
end;

Function Instr(const Source,Sub: String;const StartPos: Byte): Byte; Assembler;
asm
{$ifopt G+}
 push bp                  { save frame pointer }
{$endif}
 push ds
 cld
 les  di,Source
 lds  si,Sub
 xor  bx,bx
 or   bl,startpos
 jz   @nochange           { don't change when zero }
 dec  bl
@nochange:
 lodsw                    { AL = length source, AH = 1. char in source }
 sub  al,1                { length source - 1 }
 jb   @notfound1
 mov  cl,es:[di]          { length(sub) }
 mov  ch,cl
 sub  cl,al               { length(source)-length(sub)+1 }
 jbe  @notfound1          { if <= 0 then exit }
 sub  cl,bl               { remaining length }
 jbe  @notfound1          { if <= 0 then exit }
 inc  di                  { 1:st char }
 add  di,bx               { start position in sub }
 xchg ch,bh
 xchg al,ah
 mov  bp,si
@search1:
 repnz scasb              { scan source for 1st char of substring }
 jnz  @notfound2          { if not found, exit }
 mov  bl,cl               { save remaining length source }
 mov  cl,ah               { char still to compare in substring }
 mov  dx,di               { get offset into source }
 repz cmpsb               { compare CX chars of source and substring }
 mov  SI,BP               { offset into substring }
 mov  DI,DX               { offset into source }
 mov  CL,BL               { remaining length source }
 jnz  @search1            { continue search if no complete match }
 sub  bh,bl               { pos = length source - remaining len source}
 sub  bh,ah               { pos = l. source - rem. l. source - l. substring}
@NotFound1:
 mov  ch,bh               { pos }
@NotFound2:
 mov  al,ch
 xor  ah,ah
 pop  ds
{$IFOPT G+}
 pop  bp                  { restore frame pointer }
{$ENDIF}
end;

Procedure PadRight_P(var S: String;const Count,PadChar: Byte); Assembler;
asm
 les  di,S
 mov  dl,byte ptr es:[di]
 xor  ch,ch
 mov  cl,count
 jcxz @done
 cmp  dl,cl
 jg   @done
 mov  bx,di
 xor  dh,dh
 add  di,dx
 sub  cl,dl
 inc  di
 cld
 mov  al,padchar
 rep  stosb
 mov  di,bx
 mov  cl,count
@done:
 mov  es:[di],cl
@exit:
end;

Function PadRight_F(const S: String;const Count,PadChar: Byte): String; Assembler;
asm
 cld
 push ds
 les  di,@result
 lds  si,S
 mov  bx,di
 mov  cl,byte ptr ds:[si]
 inc  cl
 rep  movsb
 mov  di,bx
 mov  dl,byte ptr es:[di]
 xor  ch,ch
 mov  cl,count
 jcxz @done
 cmp  dl,cl
 jg   @done
 xor  dh,dh
 add  di,dx
 sub  cl,dl
 inc  di
 cld
 mov  al,padchar
 rep  stosb
 mov  di,bx
 mov  cl,count
@done:
 mov  es:[di],cl
 pop  ds
end;

Function StripFromRight(const S: String;const BreakChar: Char): String; Assembler;
asm
 push ds
 les  di,S
 xor  ch,ch
 mov  cl,byte ptr[di]   { get length of string }
 mov  bl,cl
 add  di,cx             { move to end of string }
 std                    { scan backwards }
 mov  al,BreakChar
 repne scasb
 cld
 jcxz @notfound         { not found = copy whole string }
 inc  cl
 jmp  @doit             { found     = copy part-of-string }
@notfound:
 mov  cl,bl
@doit:
 lds  si,S
 les  di,@result
 mov  al,cl
 stosb
 inc  si
 rep  movsb
@exit:
 pop  ds
end;


END.
